#! /usr/bin/python
"""
Create block wise subnet visual overview Excel file.
Subnet repository is used as a data basis for this.

This script will only work with IPv4 addresses.

Copyright (c) 2017 by Kostis Netzwerkberatung
Talstr. 25, D-63322 Roedermark, Tel. +49 6074 881056
kosta@kostis.de (Konstantinos Kostis), http://www.kostis.de/

You may use this script free of charge at your own risk.
"""

import getopt
import sys
import os
from pathlib import Path
import time
import ipaddress

from openpyxl import Workbook, load_workbook
from openpyxl.cell import Cell
from openpyxl.comments import Comment
from openpyxl.styles import colors, Color, PatternFill, Border, Side, Alignment, Font, Fill

""" Python SCRIPT (base)name """
SCRIPT = os.path.basename(sys.argv[0])
VERSION = 'V0.10 (2017-07-07)'

DEBUG = 0						# value != 0 will enable debugging messages to sys.stderr

""" IPv4 subnet specific values """
IPV4_MAX_BITS = 32				# number of bits for an IPv4Address

""" IPv4 overview values """
blocksize = 22					# default value, changed by option
subnetsize = 29					# default value, changed by option

""" Helps to convert from integer to capital lett """
ASC_OFFSET = 64					# last ASCII character before capital letters

""" Border style and color """
BORDER_STYLE = 'medium'
BORDER_COLOR = colors.BLACK

""" RBG colors for overview file """
COLOR_FREE = 'FF3FFF3F'
COLOR_USED = 'FF00BFFF'
COLOR_SUMMARIZED = 'FF007F7F'

COLOR_HEADER_FONT = colors.WHITE
COLOR_HEADER_FILL = colors.BLACK

""" Data input file (default values, changed by options) """
repository = 'kostis_vlan_repository.xlsx'
worksheet = 'VLANs'
colname_vlan = 'VLAN'
colname_vlan_name = 'VLAN_name'
colname_ipnetaddr = 'ipnetaddr'
colname_ipmask = 'ipmask'

""" Overview file """
overview = 'kostis_subnet_overview.xlsx'	# default value, changed by option
COLUMN_WIDTH = 14				# works for default font and IPv4 address
COLUMN_WIDTH_LOG = 100

""" Comment for summarized subnets """
SUMMARIZED = 'SUMMARIZED'		# this string must not contain ';'

""" Global dictionaries read by read_repository() """
sizes = {}						# keep track of number of subnets per ipbitmask size
blocks = {}						# number of subnets within blocks
subnets = {}					# subnets read

def WhatAmI():
	""" Display information about this script (to STDERR) """
	print('Copyright (c) 2017 by Kostis Netzwerkberatung', file=sys.stderr)
	print('Written by Konstantinos Kostis (kosta@kostis.net)', file=sys.stderr)
	print('Talstr. 25, 63322 Roedermark, Germany', file=sys.stderr)
	print(file=sys.stderr)
	print(SCRIPT, VERSION, file=sys.stderr)
	print(file=sys.stderr)
	print('You may use this script free of charge at your own risk.', file=sys.stderr)
	print(file=sys.stderr)
	
def Syntax():
	""" Say what we are and display script command line syntax """
	WhatAmI()
	print('Syntax:', SCRIPT, '[options]', file=sys.stderr)
	print(file=sys.stderr)
	print('options are:', file=sys.stderr)
	print(file=sys.stderr)
	print('  -h --help', file=sys.stderr)
	print('  -b --blocksize=n', file=sys.stderr)
	print('  -s --subnetsize=n', file=sys.stderr)
	print('  -r --repository=filename', file=sys.stderr)
	print('  -w --worksheet=worksheet_name', file=sys.stderr)
	print('  -o --overview=filename', file=sys.stderr)
	print('  -v --vlan=column_name', file=sys.stderr)
	print('  -n --vlan_name=column_name', file=sys.stderr)
	print('  -i --ipnetaddr=column_name', file=sys.stderr)
	print('  -m --mask=column_name', file=sys.stderr)

def print_log(wslog, logcol, logrow, logtxt):
	"""
	Print text and log text to Excel worksheet.
	returns	next logrow
	"""
	print(logtxt)
	cellname = logcol + str(logrow)
	wslog[cellname] = logtxt
	logrow += 1
	return (logrow)

def get_net_address (ipaddr, ipbitmask):
	"""
	Return IPv4Network address for a given IPv4Address and ipbitmask.
	ipaddr must be of type str, ipbitmask must be of type int.
	Returns IPv4Network
	"""
	ipnet = ipaddress.IPv4Network(ipaddr + '/' + str(ipbitmask), strict=False)
	return(ipnet)

def create_subnet_overview(ws, ipsubnet):
	"""
	Create Excel file containg subnet overiew.
	ipsubnet must be of tyoe str
	"""
	
	global blocks
	global subnets
	global blocksize
	global subnetsize
	
	ipnetaddr, ipbitmask = ipsubnet.split('/')
	ipbitmask = int(ipbitmask)

	""" Worksheet names cannot contain '/', so we use '-' instead """
	ws.title = str(ipnetaddr) + '-' + str(ipbitmask)
	ws.freeze_panes = 'A2'

	""" Variables for subnet size statistics """
	smallest = subnetsize
	biggest = ipbitmask
	
	ipnet = ipaddress.IPv4Network(ipsubnet)
	ipnetstart = ipnet.network_address
	num_subnets = blocks[ipnet]

	""" Header format stuff """
	HeaderFontObj = Font(bold=True, color=COLOR_HEADER_FONT)
	HeaderFillObj = PatternFill(fill_type='solid', fgColor=COLOR_HEADER_FILL)

	""" Border format stuff """
	borderLeftRright = Border(\
		left=Side(border_style=BORDER_STYLE, color=BORDER_COLOR),
		right=Side(border_style=BORDER_STYLE, color=BORDER_COLOR))
	borderTop = Border(\
		left=Side(border_style=BORDER_STYLE, color=BORDER_COLOR), \
		right=Side(border_style=BORDER_STYLE, color=BORDER_COLOR), \
		top=Side(border_style=BORDER_STYLE, color=BORDER_COLOR))
	borderBottom = Border(\
		left=Side(border_style=BORDER_STYLE, color=BORDER_COLOR), \
		right=Side(border_style=BORDER_STYLE, color=BORDER_COLOR), \
		bottom=Side(border_style=BORDER_STYLE, color=BORDER_COLOR))

	""" Calculate step and number of rows """
	step = 2**(IPV4_MAX_BITS - smallest)	# initial value
	rows = 2**(IPV4_MAX_BITS - biggest) // step

	cellname = chr(ASC_OFFSET + smallest - biggest + 1) + '1'
	ws[cellname] = '#=' + str(num_subnets)
	cell = ws[cellname]
	cell.font = HeaderFontObj
	cell.fill = HeaderFillObj
	
	increment = 1	# initial value
	for bitmask in range(smallest, biggest, -1):
		""" Create header """
		column = bitmask - biggest
		colname = chr(ASC_OFFSET + column)	# works for A-Z
		cellname = colname + '1'			# 1st row is headers row
		ws[cellname] = '/' + str(bitmask)
		cell = ws[cellname]
		cell.font = HeaderFontObj
		cell.fill = HeaderFillObj
	
		""" Set colums width """
		ws.column_dimensions[colname].width = COLUMN_WIDTH
	
		""" Draw borders left and right (mind the order) """
		for row in range(rows):
			""" Row starts at 1, skip 1st row, so add 2 """
			cellname = chr(ASC_OFFSET + column) + str(row + 2)
			cell = ws[cellname]
			cell.border = borderLeftRright

		""" Paint in ipnetaddr """
		curip = ipnetstart
		for row in range(0, rows, increment):
			""" Default paint color for a cell is 'free' """
			CellComment = None
			CellFillObj = PatternFill(fill_type='solid', fgColor=COLOR_FREE)
			
			""" Figure out if subnet is used """
			key = str(curip) + '/' + str(bitmask)
			if key in subnets:
				""" mark subnet as 'used' """
				if DEBUG:
					print('    marking ', key, sep='')
				CellFillObj = PatternFill(fill_type="solid", fgColor=COLOR_USED)
				
				""" Split lines with vlan information """
				comment = subnets[key].replace(';', '\n')
				CellComment = Comment(comment, SCRIPT)
				
				""" use a different color if subnet is summarized """
				if comment.startswith(SUMMARIZED):
					CellFillObj = PatternFill(fill_type="solid", fgColor=COLOR_SUMMARIZED)
				
			""" Set value for cell and paint it """
			colname = chr(ASC_OFFSET + column)
			cellname = colname + str(row + 2)
			ws[cellname] = str(curip)
			cell = ws[cellname]
			cell.border = borderTop
			cell.fill = CellFillObj
			
			if CellComment is not None:
				cell.comment = CellComment

			""" Paint the rest of the subnet in the same color """
			for rest in range(row + 1, row + increment):
				cellname = colname + str(rest + 2)
				ws[cellname].fill = CellFillObj

			curip += step			# next subnet
	
		""" Take care of bottom line border """
		cellname = chr(ASC_OFFSET + column) + str(rows + 1)
		cell = ws[cellname]
		cell.border = borderBottom
	
		""" Next subnet mask """
		step *= 2
		increment *= 2

def get_indexed_value(ws, row, col_name):
	"""
	Return value on an indexed column.
	Ignore data formatted with strikethrough.
	"""
	cellname = col_name + str(row)
	value = ws[cellname].value
	if ws[cellname].font.strikethrough:
		value = None
	if value is None:
		value = ''
	return(value)

def read_repository(filename, worksheet):
	""""
	Read subnet information from repository file.
	Data read is stored in global sizes and blocks.
	Return number of subnets read.
	"""
	global sizes
	global blocks
	global blocksize
	global subnetsize
	
	n_subnets = 0
	
	wb = load_workbook(filename)
	ws = wb.get_sheet_by_name(worksheet)

	min_row, max_row = ws.min_row, ws.max_row
	min_col, max_col = ws.min_column, ws.max_column
	
	""" Read headers """
	fields = {}
	for col in range(min_col, max_col + 1):
		col_name = chr(ASC_OFFSET + col)
		cellname = col_name + str(min_row)
		value = ws[cellname].value
		fields[value] = col_name
	
	""" Initialize sizes dictionary """
	sizes = {}
	for size in range(0, IPV4_MAX_BITS + 1):
		sizes[str(size)] = 0		# contains number of subnet of given size
	
	""" Read data (skip header) """
	for row in range(min_row + 1, max_row + 1):
		""" get VLAN information """
		VLAN = get_indexed_value(ws, row, fields[colname_vlan])
		VLAN_name = get_indexed_value(ws, row, fields[colname_vlan_name])

		""" Skip any invalid VLAN IDs and empty VLAN names """
		try:
			vcheck = int(VLAN)
		except Exception:
			continue
		else:
			if vcheck < 1 or vcheck > 4095:
				continue
		if VLAN_name == '':
			continue

		"""
		Get subnet information.
		Note that there are two columns named 'mask' - either will do.
		"""
		ipnetaddr = get_indexed_value(ws, row, fields[colname_ipnetaddr])
		ipmask = get_indexed_value(ws, row, fields[colname_ipmask])
		
		if ipnetaddr == '' or ipmask == '':
			continue
			
		ipsubnet = ipnetaddr + '/' + ipmask
		
		ipnet = None
		try:
			ipnet = ipaddress.IPv4Network(ipsubnet)
		except Exception as ERR_MSG:
			continue
		
		if ipnet is not None:
			""" str(ipnet) implies conversion to bitmask format """
			ipnetaddr, ipbitmask = str(ipnet).split('/')
			sizes[ipbitmask] += 1
			
			""" Subnets greater or equal blocksize need no overview """
			if int(ipbitmask) <= blocksize:
				if (DEBUG):
					print('### DEBUG: ignoring ', str(ipnet), ' (size >= /', blocksize, ')', \
						file=sys.stderr, sep='')
				continue
				
			supernet = get_net_address(ipnetaddr, blocksize)
			if supernet in blocks:
				blocks[supernet] += 1
			else:
				blocks[supernet] = 1
			
			addcomment = str(VLAN) + ' ' + VLAN_name
			if int(ipbitmask) <= subnetsize:
				subnets[str(ipnet)] = addcomment
			else:
				""" Summarize subnets below SUBNET_SIZE """
				subnet = get_net_address(ipnetaddr, subnetsize)
				strsubnet = str(subnet)
				if strsubnet in subnets:
					subnets[strsubnet] += ';' + addcomment
				else:
					subnets[strsubnet] = SUMMARIZED + ';' + addcomment
				if (DEBUG):
					print('### DEBUG ', SCRIPT, ': modified ', str(ipnet), ' to ', subnet, file=sys.stderr, sep='')
			
			n_subnets += 1
	
	return (n_subnets)

def main():
	global blocksize, subnetsize
	global repository, overview
	global worksheet
	global colname_vlan, colname_vlan_name, colname_ipnetaddr, colname_ipmask
	
	""" Check syntax and get options """
	try:
		opts, args = getopt.getopt(sys.argv[1:], \
			'hb:s:r:o:v:n:i:m:', \
			['help', 'blocksize=', 'subnetsize=', 'repository=', 'overview=', 'vlan=', 'vlan_name=', 'ipnetaddr=', 'msak='])
	except getopt.GetoptError as err:
		Syntax()
		return(1)
	
	for o, a in opts:
		if o in ('-h', '--help'):
			Syntax()
			return(1)
		elif o in ('-o', '--overview'):
			overview = a
		elif o in ('-r', '--repository'):
			repository = a
		elif o in ('-b', '--blocksize'):
			blocksize = int(a)
		elif o in ('-s', '--subnetsize'):
			blocksize = int(a)
		elif o in ('-w', '--worksheet'):
			worksheet = a
		elif o in ('-v', '--vlan'):
			colname_vlan = a
		elif o in ('-n', '--vlan_name'):
			colname_vlan_name = a
		elif o in ('-i', '--ipnetaddr'):
			colname_ipnetaddr = a
		elif o in ('-m', '--mask'):
			colname_ipmask = a
		else:
			assert False, 'unhandled option'

	""" Figure out file date/time """
	epoch = os.path.getmtime(repository)
	datadate = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(epoch))
	
	print('reading Excel workbook ', repository, sep='')
	n_subnets = read_repository(repository, worksheet)
	
	wb = Workbook()
	wslog = wb.active
	wslog.title = 'log'

	""" Worksheet log starts in first column and row """
	logcol = 'A'
	logrow = 1

	""" Set column width """
	wslog.column_dimensions[logcol].width = COLUMN_WIDTH_LOG

	""" Begin log to worksheet and sys.stdout """
	logrow = print_log(wslog, logcol, logrow, SCRIPT + ' ' + VERSION)
	logrow = print_log(wslog, logcol, logrow, 'repository ' + repository + ', worksheet ' + worksheet)
	logrow = print_log(wslog, logcol, logrow, 'repository date/time ' + datadate)
	logrow = print_log(wslog, logcol, logrow, 'blocksize=' + str(blocksize) + ', subnetsize=' + str(subnetsize))
	
	""" Display subnet size summary """
	total = 0
	for size in sizes:
		num = sizes[size]
		if (num > 0):
			total += num
			logtxt = '/' + str(size) + ' has ' + str(num) + ' entries'
			if int(size) <= blocksize:
				logtxt += ' (ignored)'
			logrow = print_log(wslog, logcol, logrow, logtxt)

	""" More subnet statistics """
	logrow = print_log(wslog, logcol, logrow, 'total number of relevant subnets is ' + str(total))
	logrow = print_log(wslog, logcol, logrow, 'number of blocks is ' + str(len(blocks)))
	
	""" Create overview block by block """
	for block in sorted(blocks):
		logtxt = 'creating overview for block ' + str(block) + ' with ' + str(blocks[block]) + ' subnet(s)'
		logrow = print_log(wslog, logcol, logrow, logtxt)
		ws = wb.create_sheet('SPAM')	# will be renamed by create_subnet_overview()
		create_subnet_overview(ws, str(block))

	print('writing Excel workbook ', overview, sep='')
	wb.save(overview)
	
	return(0)

if __name__ == '__main__':
	rc = main()
	sys.exit(rc)
